//
// Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
// Ministerpräsidenten des Landes Schleswig-Holstein
// Staatskanzlei
// Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
//
// Lizenziert unter der EUPL, Version 1.2 oder - sobald
// diese von der Europäischen Kommission genehmigt wurden -
// Folgeversionen der EUPL ("Lizenz");
// Sie dürfen dieses Werk ausschließlich gemäß
// dieser Lizenz nutzen.
// Eine Kopie der Lizenz finden Sie hier:
//
// https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
//
// Sofern nicht durch anwendbare Rechtsvorschriften
// gefordert oder in schriftlicher Form vereinbart, wird
// die unter der Lizenz verbreitete Software "so wie sie
// ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
// ausdrücklich oder stillschweigend - verbreitet.
// Die sprachspezifischen Genehmigungen und Beschränkungen
// unter der Lizenz sind dem Lizenztext zu entnehmen.
//

package mock

import (
	pb "antragsraum-proxy/gen/go"
	"context"
	"github.com/stretchr/testify/assert"
	"google.golang.org/grpc"
	"google.golang.org/grpc/credentials/insecure"
	"net"
	"testing"
	"time"
)

func TestStartGrpcServer(t *testing.T) {
	t.Run("should start gRPC server", func(t *testing.T) {
		SetUpGrpcServer()

		conn, err := net.DialTimeout("tcp", "localhost:50052", 2*time.Second)

		assert.NoError(t, err)

		conn.Close()
	})

	t.Run("TestAntragraumEndpoints", func(t *testing.T) {
		setUpGrpcEnv := func() (pb.AntragraumServiceClient, func()) {
			SetUpGrpcServer()

			conn, err := grpc.NewClient("localhost:50052", grpc.WithTransportCredentials(insecure.NewCredentials()))
			assert.NoError(t, err)

			cleanup := func() {
				conn.Close()
			}
			client := pb.NewAntragraumServiceClient(conn)

			return client, cleanup
		}

		t.Run("TestFindRueckfragen", func(t *testing.T) {
			t.Run("should have no error", func(t *testing.T) {
				client, cleanUp := setUpGrpcEnv()
				defer cleanUp()

				resp, err := client.FindRueckfragen(context.Background(), &pb.GrpcFindRueckfragenRequest{PostfachId: "testPostfachId"})

				assert.NoError(t, err)
				assert.NotNil(t, resp)
			})

			t.Run("should have error", func(t *testing.T) {
				client, cleanUp := setUpGrpcEnv()
				defer cleanUp()

				resp, err := client.FindRueckfragen(context.Background(), &pb.GrpcFindRueckfragenRequest{})

				assert.Error(t, err)
				assert.Nil(t, resp)
			})
		})

		t.Run("TestGetRueckfrage", func(t *testing.T) {
			t.Run("should have no error", func(t *testing.T) {
				client, cleanUp := setUpGrpcEnv()
				defer cleanUp()

				resp, err := client.GetRueckfrage(context.Background(), &pb.GrpcGetRueckfrageRequest{Id: "testId"})

				assert.NoError(t, err)
				assert.NotNil(t, resp)
			})

			t.Run("should have error", func(t *testing.T) {
				client, cleanUp := setUpGrpcEnv()
				defer cleanUp()

				resp, err := client.GetRueckfrage(context.Background(), &pb.GrpcGetRueckfrageRequest{})

				assert.Error(t, err)
				assert.Nil(t, resp)
			})
		})

		t.Run("TestSendRueckfrageAnswer", func(t *testing.T) {
			client, cleanUp := setUpGrpcEnv()
			defer cleanUp()

			resp, err := client.SendRueckfrageAnswer(context.Background(), &pb.GrpcSendRueckfrageAnswerRequest{})

			assert.NoError(t, err)
			assert.NotNil(t, resp)
		})

		t.Run("TestGetAttachmentContent", func(t *testing.T) {
			client, cleanUp := setUpGrpcEnv()
			defer cleanUp()

			resp, err := client.GetAttachmentContent(context.Background(), &pb.GrpcGetAttachmentContentRequest{})

			assert.NoError(t, err)
			assert.NotNil(t, resp)
		})

		t.Run("TestGetAttachmentMetadata", func(t *testing.T) {
			t.Run("should have no error", func(t *testing.T) {
				client, cleanUp := setUpGrpcEnv()
				defer cleanUp()

				resp, err := client.GetAttachmentMetadata(context.Background(), &pb.GrpcGetAttachmentMetadataRequest{FileId: "testFileId"})

				assert.NoError(t, err)
				assert.NotNil(t, resp)
			})

			t.Run("should have error", func(t *testing.T) {
				client, cleanUp := setUpGrpcEnv()
				defer cleanUp()

				resp, err := client.GetAttachmentMetadata(context.Background(), &pb.GrpcGetAttachmentMetadataRequest{})

				assert.Error(t, err)
				assert.Nil(t, resp)
			})

		})
	})

	t.Run("TestCommandEndpoints", func(t *testing.T) {
		setUpGrpcEnv := func() (pb.CommandServiceClient, func()) {
			SetUpGrpcServer()

			conn, err := grpc.NewClient("localhost:50052", grpc.WithTransportCredentials(insecure.NewCredentials()))
			assert.NoError(t, err)

			cleanup := func() {
				conn.Close()
			}
			client := pb.NewCommandServiceClient(conn)

			return client, cleanup
		}

		t.Run("should have no error", func(t *testing.T) {
			client, cleanUp := setUpGrpcEnv()
			defer cleanUp()

			resp, err := client.GetCommand(context.Background(), &pb.GrpcGetCommandRequest{Id: "testId"})

			assert.NoError(t, err)
			assert.NotNil(t, resp)
		})

		t.Run("should have error", func(t *testing.T) {
			client, cleanUp := setUpGrpcEnv()
			defer cleanUp()

			resp, err := client.GetCommand(context.Background(), &pb.GrpcGetCommandRequest{})

			assert.Error(t, err)
			assert.Nil(t, resp)
		})
	})

	t.Run("TestBinaryFileEndpoints", func(t *testing.T) {
		setUpGrpcEnv := func() (pb.BinaryFileServiceClient, func()) {
			SetUpGrpcServer()

			conn, err := grpc.NewClient("localhost:50052", grpc.WithTransportCredentials(insecure.NewCredentials()))
			assert.NoError(t, err)

			cleanup := func() {
				conn.Close()
			}
			client := pb.NewBinaryFileServiceClient(conn)

			return client, cleanup
		}

		t.Run("TestUploadBinaryFileAsStream", func(t *testing.T) {
			client, cleanUp := setUpGrpcEnv()
			defer cleanUp()

			resp, err := client.UploadBinaryFileAsStream(context.Background())

			assert.NoError(t, err)
			assert.NotNil(t, resp)
		})
	})
}
